;| <--------------------------------------------------------------------------

Function: Mtanlc, Mtancl, and Tanlc

Moves a line tangent to a circle or a circle tangent to the line depending
upon the command chosen. MtanLC is line to circle and mtanCL is circle to
line. Select the point on the line that is to be tangent to the circle, then
select the circle. The line will be moved or copied, users choice by prompt,
to the circle or vise versa. The side of the line or circle that the tangent
lies will be arbitrary because it depends upon the direction that the line
was drawn originally. Therefore, after the tangent move/copy is completed,
the user will be prompted to leave it as is or to change sides.

Arbitrary input is allowed for the selection of the tangent point. Therefore
a transparent Lisp or CAL expression may be entered for this prompt and will
be honored as long as its return is a point.

Works with arcs as if they were circles. The arc may not have a tangent point
to the line, ie. they do not touch, but if extended, the arc will go to the
tangent point.

//LL 4/5/95
Modified to work with straight polyline segments.

Written by Larry Leuallen 4/5/95
(C) Copyright 1995 Larry Leuallen
All rights reserved.

LARRY LEUALLEN PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. LARRY
LEUALLEN SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR USE. LARRY LEUALLEN DOES NOT WARRANT THAT THE
OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.

--------------------------------------------------------------------------> |;
;;;--------------------------------------------------------
;;; MAIN FUNCTION DEFINITIONS

;;; Two separate functions.

(defun c:mtanlc ()                    ;MoveTangentLinetoCircle
  (mtan "moveline")
) ;end defun

(defun c:mtancl ()                    ;MoveTangentCircletoLine
  (mtan "movecircle")
) ;end defun

;;; Combined functions. Comment out above or below.

(defun c:tanlc (/ key)
  (initget 1 "Circle Line")
  (setq key (getkword (strcat "\nMove/copy Circle tangent to line or "
    "Line tangent to circle C/L: ")))
  (cond
    ( (= key "Circle")
      (mtan "movecircle")
    )
    ( T
      (mtan "moveline")
    )
  ) ; end cond.
) ;end defun

;;;--------------------------------------------------------
;;; MTAN
;;; Main subroutine. Mode is "moveline" or "movecircle".

(defun mtan (mode / oldcmd oldosm olderr newerr ename ln elist ercmd
                    bp cn cen rad mp pmpt e1 e2 key1 key2 kind lename
            )
(defun newerr (NE)
  (cond
    ( (= NE "quit / exit abort")
      (princ "User ")
    )
    ( (/= NE "Function cancelled")
      (princ (strcat "\nError: " NE))
    )
  )
  ;; If a command is active, cancel it. Loop until no command active.
  (while (> (getvar "cmdactive") 0)
    (command)
  ) ;end while
  ;; If a group is active end it and undo that group.
  (cond
    ( (= (logand 8 (getvar "undoctl")) 8)
      (command "._undo" "e")
      (command "._undo" 1)
    )
  ) ; end cond.
  (if ercmd
    (mapcar
      '(lambda (x)
        (eval x)
      ) ;end lambda
      ercmd
    ) ;end mapcar
  )
  (setvar "CMDECHO" cmde)
  (setvar "OSMODE" oldosm)
  (setq *error* olderr)
  (princ)
)
  (setq olderr *error* *error* newerr)
  (setq oldcmd (getvar "cmdecho"))
  (setvar "cmdecho" 0)
  (setq oldosm (getvar "osmode"))
  (setvar "osmode" 0)
  ;; Start a group so that the whole operation can be undone with
  ;; one undo command. If a group is active, starting a new group
  ;; will end it.
  (command "._undo" "g")
  ;; Set prompts for different modes.
  (cond
    ( (= mode "moveline")
      (prompt "\nMoves or copies a LINE/PLINE tangent to a CIRCLE.  ")
    )
    ( T
      (prompt "\nMoves or copies a CIRCLE tangent to a LINE/PLINE.  ")
    )
  ) ; end cond.
  (while (not ln)
    ;; Select line or polyline.
    (setq ln (entsel "\nSelect line or pline:  "))
    (cond
      ( ln
        (setq kind (cdr (assoc 0 (entget (car ln)))))
        (setq lename (cdr (assoc -1 (entget (car ln)))))
        (cond
          ( (not  (or  (= kind "LINE")
                       (= kind "POLYLINE")
                  )
            )
            (setq ln (entsel "\nSelect line or pline:  "))
          )
          ( T
            (redraw lename 3)
            (setq ercmd (list '(redraw lename 4)))
          )
        ) ; end cond.
      )
      ( T
        (setq ln (entsel "\nSelect line or pline:  "))
      )
    ) ; end cond.
  ) ;end while
  ;; Allow arbitrary input for tangent point.
  (while (not bp)
    (initget 128)
    (cond
      ( (= kind "LINE")
        (setq bp (getpoint "\nTangent point on line: "))
      )
      ( T
        (setq bp (getpoint "\nTangent point on straight segment pline: "))
      )
    ) ; end cond.
    (cond
      ( (= (type bp) 'STR)
        ;; If string evaluates to a point, loop will end. Else nil
        ;; returned and loop continues.
        (setq bp (mtan_esp bp))
        ;; Warn that wrong data type returned. Will loop back for
        ;; data reselection.
        (cond
          ( (not bp)
            (prompt "\nEvaluation did not return a point. ")
          )
        ) ; end cond.
      )
    ) ; end cond.
    (cond
      ( (and bp (setq bp (osnap bp "nea")))
        (cond
          ( (not (setq ln (mt_is_line bp)))
            (prompt (strcat "\nTangent point must be on "
              "straight pline segment. "))
            (setq bp nil)
          )
         )
         (cond
          ( (not (equal lename (caddr ln)))
            (prompt (strcat "\nTangent point must be on "
              "originally selected entity. "))
            (setq bp nil)
          )
        )
      )
      ( T
        (prompt "\nPoint not on entity.  ")
      )
    ) ; end cond.
  ) ;end while
  (setq e1 (car ln) e2 (cadr ln) ln (caddr ln))
  ;; Select tangent circle.
  (while (not (and (setq ename (entsel "\nTangent circle/arc: "))
                   (setq cn (car ename))
                   (setq elist (entget cn))
                   (or
                    (= "CIRCLE" (cdr (assoc 0 elist)))
                    (= "ARC" (cdr (assoc 0 elist)))
                   )
              )
         )
    (prompt "\nCircle/arc not selected! ")
  )
  (redraw cn 3)
  (setq ercmd (list '(redraw lename 4) '(redraw cn 4)))
  ;; Set correct entity to move or copy depending upon mode.
  (cond
    ( (= mode "movecircle")
      (setq ln cn)
    )
  ) ; end cond.
  ;; Get center and radius of circle.
  (setq cen (cdr (assoc 10 elist)))
  (setq rad (cdr (assoc 40 elist)))
  ;; Get the basepoint to make the move or copy by.
  (setq mp (get_bp e1 e2 rad))
  ;; Set the prompts for default move or copy based on state of
  ;; global variable.
  (cond
    ( (not #mtanlc)
      (setq pmpt "\nMove/Copy line <M>/C: ")
    )
    ( T
      (cond
        ( (= #mtanlc "Copy")
          (setq pmpt "\nMove/Copy line M/<C>: ")
        )
        ( T
          (setq pmpt "\nMove/Copy line <M>/C: ")
        )
      ) ; end cond.
    )
  ) ; end cond.
  ;; Select the move or copy option.
  (initget "Move Copy  ")
  (setq key1 (getkword pmpt))
  (cond
    ( (= mode "moveline")
      ;; The line is moved or copied.
      (cond
        ( (or (not key1) (= key1 "") (= key1 "Copy"))
          (command "._copy" ln "" mp cen)
          (redraw cn 4)
          (setq ercmd nil)
          (setq #mtanlc "Copy")
        )
        ( T
          (command "._move" ln "" mp cen)
          (redraw cn 4)
          (setq ercmd nil)
          (setq #mtanlc "Move")
        )
      ) ; end cond.
      ;; Flip position option.
      (do_flip "Line")
    )
    ( T
      ;; The circle is moved or copied.
      (cond
        ( (or (not key1) (= key1 "") (= key1 "Copy"))
          (command "._copy" ln "" cen mp)
          (redraw lename 4)
          (setq ercmd nil)
          (setq #mtanlc "Copy")
        )
        ( T
          (command "._move" ln "" cen mp)
          (redraw lename 4)
          (setq ercmd nil)
          (setq #mtanlc "Move")
        )
      ) ; end cond.
      (do_flip "Circle")
    )
  ) ; end cond.
  (setvar "cmdecho" oldcmd)
  (setvar "osmode" oldosm)
  ;; End the group.
  (command "._undo" "e")
  (setq *error* olderr)
  (princ)
) ;end defun

;;;--------------------------------------------------------
;;; GET_BP
;;; Gets the basepoint to make the move/copy from.

(defun get_bp (e1 e2 rad / mp)
  (setq mp (mtan_norln e1 e2))
  (setq mp (mapcar '* mp (list rad rad 0.0)))
  (mapcar '+ bp mp)
) ;end defun

;;;--------------------------------------------------------
;;; DO_FLIP
;;; Flips the line or circle to the other side.

(defun do_flip (mode / key2 temp)
  (initget "Yes No  ")
  (setq key2 (getkword "\nOther side? Y/<N>: "))
  (cond
    ( (not (or (= key2 "") (not key2) (= key2 "No")))
      ;;---------------------------------
      ;; Undo it and redo with normal direction reversed.
      (command "._undo" 1)
      (setq mp (get_bp e2 e1 rad))
      ;;---------------------------------
      ;; Reverse variables if circle mode
      (if (= mode "Circle")
        (progn
          (setq temp mp mp cen cen temp)
        ) ;end progn
      )
      (cond
        ( (or (not key1) (= key1 "") (= key1 "Copy"))
          (command "._copy" ln "" mp cen)
          (setq #mtanlc "Copy")
        )
        ( T
          (command "._move" ln "" mp cen)
          (setq #mtanlc "Move")
        )
      ) ; end cond.
    )
  ) ; end cond.
) ;end defun

;;;--------------------------------------------------------//LL 4/5/95
;;; MT_IS_LINE
;;; Given a point, returns endpoints of line and entity name in list if
;;; line, vertex endpoints and parent entity name if straight polyline
;;; segment, or else nil.

(defun MT_IS_LINE (pt / ss1 n_pt ename)
  (setq ss1 (ssget pt))
  (cond
    ( ss1
      (setq ss1 (entget (ssname ss1 0)))
      (setq ename (cdr (assoc -1 ss1)))
      (cond
        ( (= (cdr (assoc 0 ss1)) "LINE")
          (setq ss1 (list (cdr (assoc 10 ss1)) (cdr (assoc 11 ss1)) ename))
        )
        ( (= (cdr (assoc 0 ss1)) "POLYLINE")
          (setq ss1 (entget (car (nentselp pt))))    ;segment at pt, ename
          (cond
            ( (= (cdr (assoc 42 ss1)) 0.0)  ;bulge zero is straight line
              (setq n_pt (cdr (assoc 10 ss1))
                    ss1 (entget (entnext (cdr (assoc -1 ss1))))
                    ss1 (list n_pt (cdr (assoc 10 ss1)) ename)
              )
            )
            ( T                           ;not straight
              (setq ss1 nil)
            )
          ) ; end cond.
        )
        ( T                               ;not line or polyline
          (setq ss1 nil)
        )
      ) ; end cond.
    )
  ) ; end cond.
  ss1
)

;;;--------------------------------------------------------
;;; MTAN_NORLN
;;; Unit normal vector to a line. Same as CAL function nor(p1,p2).

(defun mtan_norln (p1 p2 / point)
  (setq point (polar p1 (+ (angle p1 p2) (/ pi 2)) 1.0))
  (mapcar '- point p1)
)

;;;--------------------------------------------------------
;;; EVAL_STR_POINT
;;; If the evaluation is not nil the function returns the evaluation as a
;;; point if it were a point, else returns nil. The point can be a double
;;; or a tripple.

(defun MTAN_ESP (str)
  (setq str (strcase str))
  (cond
    ( (or (= str "'CAL" ) (= str "CAL") (= str "'_CAL" ))
      (setq str "(c:cal)")
    )
  ) ; end cond.
  (cond
    ( (setq str (eval (read str)))
      (cond
        ( (not  (and  (= (type str) 'LIST)
                      (or (and  (= (type (car str)) 'REAL)
                                (= (type (cadr str)) 'REAL)
                                (= (type (caddr str)) 'REAL)
                                (= (length str) 3)

                            ) ;end and
                          (and (= (type (car str)) 'REAL)
                                (= (type (cadr str)) 'REAL)
                                (= (length str) 2)

                          ) ;end and
                      ) ;end or
                ) ; end and
          ) ;end not
          (setq str nil)
        )
      ) ; end cond
    )
  ) ; end cond.
  str
) ;end defun

(princ)